import os
import json

from pathlib import Path

from notebook.base.handlers import APIHandler
from notebook.utils import url_path_join

from subprocess import Popen
import tornado
import uuid

import inhpc_dm.uftp_handler as uftp_handler
import inhpc_dm.tasks as tasks


class AbstractDMHandler(APIHandler):
    """ Abstract base class for dm-tool handlers """
    
    def _get_settings_path(self):
        """
            File where the current user preferences for the DM tool are stored
        """
        settings_dir = os.environ['HOME']+"/.inhpc/"
        self._assert_dir_exists(settings_dir)
        return settings_dir + "settings.json"

    def _get_mount_info_path(self):
        """
            File where the current mounts are stored
        """
        settings_dir = os.environ['HOME']+"/.inhpc/"
        self._assert_dir_exists(settings_dir)
        return settings_dir + "mounts.json"

    def _assert_dir_exists(self, name):
        try:
            target = Path(name);
            if not (target.exists() and target.is_dir()):
                target.mkdir()
        except:
            self.log.error("Cannot create directory: '%s' " % name)

    def read_mount_info(self):
        try:
            with open(self._get_mount_info_path(), 'r') as f:
                return json.load(f)
        except:
            return {}

    def store_mount_info(self, mount_info):
        with open(self._get_mount_info_path(), 'w') as f:
            f.write(json.dumps(mount_info))

    def resolve_mount_point(self, directory, mount_info=None):
        """ find and return the mount info which contains the given directory """
        if mount_info==None:
            mount_info = self.read_mount_info()
        lookup = list(Path(directory).absolute().parts)
        for id in mount_info:
            m = mount_info.get(id)
            mount_point = list(Path(m.get("mount_point")).parts)
            # check if mount point is a parent dir for our target dir
            if mount_point == lookup[0:len(mount_point)]:
                return id, m
        # nothing found
        return None, None

    def write_error(self, status_code, **kwargs):
        """  override to send back error message as JSON content """
        self.set_header("Content-Type", "application/json")
        msg = self._reason
        if "exc_info" in kwargs:
            try:
                msg = msg + " [" + str(kwargs["exc_info"][1]) + "]"
            except:
                pass
        self.finish('{"message": "%s"}' % msg)


class MountHandler(AbstractDMHandler):
    """
    REST API dealing with mount
    
    Mount information is stored in a JSON file $HOME/.inhpc/mounts.json

    """

    def force_unmount(self, mount_dir):
        child = Popen("fusermount -u %s" % mount_dir, shell=True)
        child.wait()


    @tornado.web.authenticated
    def get(self):
        """
        Get all mounts
        """
        self.finish(json.dumps({"mount_info": self.read_mount_info()}))

    @tornado.web.authenticated
    def post(self):
        """
        Create a new mount
        
        input: JSON object 
           {'protocol': 'uftp',
            'mount_point': 'local_directory',
            ... additional protocol-dependent parameters
           }
        
        Input contains also the protocol-dependent information 
        required to initiate the mount such as remote URL, remote directory etc
        """
        request_data = self.get_json_body()
        protocol = request_data.get("protocol", "uftp")
        mount_point = request_data['mount_point']
        if not mount_point.startswith("/"):
            mount_point = str(Path(mount_point).absolute())
            request_data['mount_point'] = mount_point
        self.force_unmount(mount_point)
        self._assert_dir_exists(mount_point)
        if "uftp"==protocol:
            parameters = request_data
            exit_code, error_info = uftp_handler.mount(mount_point, parameters)
            id = uuid.uuid4().hex
            mount_info = self.read_mount_info()
            parameters["protocol"] = "uftp"
            mount_info[id] = parameters
        else:
            raise Exception("No handler for protocol %s", protocol)
        result_data = {}
        if exit_code == 0:
            result_data["status"] = "OK"
            result_data["id"] = id
            self.store_mount_info(mount_info)
        else:
            result_data["status"] = "ERROR"
            result_data["exit_code"] = exit_code
            result_data["error_info"] = error_info
        self.finish(json.dumps(result_data))


class UnmountHandler(AbstractDMHandler):
    """
    REST API dealing with unmount

    """

    @tornado.web.authenticated
    def post(self):
        """
        Un-mount a directory
        
        input: JSON object 
           {
             'mount_point': 'local_directory',
           }
        """
        request_data = self.get_json_body()
        mount_point = request_data['mount_point']
        if not mount_point.startswith("/"):
            mount_point = str(Path(mount_point).absolute())
            request_data['mount_point'] = mount_point
        id, mount = self.resolve_mount_point(mount_point)
        if mount!=None:
            request_data['mount_point'] = mount["mount_point"]
            exit_code, error_info = uftp_handler.unmount(request_data)
        else:
            exit_code, error_info = -1, "Not a mounted directory"
        result_data = { "status": "OK" }
        if exit_code == 0:
            mount_info = self.read_mount_info()
            mount_info.pop(id)
            self.store_mount_info(mount_info)
        else:
            result_data["status"] = "ERROR"
            result_data["exit_code"] = exit_code
            result_data["error_info"] = error_info
        self.finish(json.dumps(result_data))

class TaskHandler(AbstractDMHandler):
    """
    REST API dealing with tasks (copy operations etc)
    """

    @tornado.web.authenticated
    def get(self):
        """
        Get tasks
        """
        self.finish(json.dumps({"tasks": [ t.json() for t in self.application.dm_task_holder.tasks]}))

    @tornado.web.authenticated
    def post(self):
        """
        Launch a new task
        
        input: JSON object 
           {
            "command": "<command_template_name or command>"
            "parameters": [...]
           }
        """
        request_data = self.get_json_body()
        cmd = request_data.get("command", "")
        args = request_data.get("parameters", {})
        if "copy"!=cmd:
            raise ValueError("Not understood: %s" % cmd)
        sources = args["sources"]
        target = args["target"]
        if len(target)==0:
            target = "."
        mount_info = self.read_mount_info()
        id_1, target_mount = self.resolve_mount_point(target, mount_info)
        id_2, source_mount = self.resolve_mount_point(sources[0], mount_info)
        if id_1==None and id_2==None:
            raise ValueError("Neither source nor target are remote")
        if id_1!=None and id_2!=None:
            raise ValueError("Cannot have both remote source and remote target")
        
        cmd = uftp_handler.prepare_data_move_operation(sources, target, mount_info)
        self.log.info("Running: %s" % cmd)
        task = tasks.Task(cmd)
        task.launch()
        self.application.dm_task_holder.add(task)
        result_data = { "status": "OK" }
        self.finish(json.dumps(result_data))


def setup_handlers(web_app, url_path):
    host_pattern = ".*$"
    base_url = web_app.settings["base_url"]
    handlers = [(url_path_join(base_url, url_path, "mount"), MountHandler),
                (url_path_join(base_url, url_path, "unmount"), UnmountHandler),
                (url_path_join(base_url, url_path, "tasks"), TaskHandler),
                ]
    web_app.dm_task_holder = tasks.TaskHolder()
    web_app.add_handlers(host_pattern, handlers)



from base64 import b64encode
from fuse import FUSE
from os import environ, getenv
from pathlib import Path, PosixPath, PurePosixPath
from pyunicore.client import Transport
from pyunicore.credentials import create_credential
from pyunicore.uftp import UFTP
from pyunicore.uftpfuse import UFTPDriver

from stat import S_ISDIR
import subprocess


def _setup_credential(credentials):
    # cleanup credential data we got from the UI side
    _user = credentials.get("username", "")
    if len(_user)==0:
        _user = None
    _password = credentials.get("password", "")
    if len(_password)==0:
        _password = None
    _token = credentials.get("token", "")
    if len(_token)==0:
        _token = None
    _identity = credentials.get("identity", "")
    if len(_identity)==0:
        _identity = None
    return create_credential(username = _user,
                             password = _password,
                             token = _token,
                             identity = _identity)


def create_uftp_handler(remote_directory, auth_url, credentials):
    """ creates a UFTP handler (not yet authenticated or connected) """
    credential = _setup_credential(credentials)
    return UFTP(Transport(credential), auth_url, remote_directory)


def run_fusedriver(host, port, pwd, mount_point, debug=False):
    cmds = ["export PYTHONPATH=%s" % environ.get("PYTHONPATH", ""),
            "export UFTP_PASSWORD=%s" % pwd,
            "python3 -m pyunicore.uftpfuse %s:%s '%s'" % (host, port, mount_point)]
    cmd = ""
    for c in cmds:
        cmd += c + u"\n"
    return _run_command(cmd)


def _resolve_mount_point(directory, mount_info):
        """ find and return the mount info which contains the given directory """
        lookup = list(Path(directory).absolute().parts)
        for id in mount_info:
            m = mount_info.get(id)
            mount_point = list(Path(m.get("mount_point")).parts)
            # check if mount point is a parent dir for our target dir
            if mount_point == lookup[0:len(mount_point)]:
                return m
        # nothing found
        return None

    
def _create_command(host, port, pwd, method, sources, target):
    cmd =  "export PYTHONPATH=%s\n" % environ.get("PYTHONPATH", "")
    cmd += "export UFTP_PASSWORD=%s\n" % pwd
    cmd += "python3 -m inhpc_dm.uftp_handler %s:%s -X %s " %(host,port, method)
    if len(sources)<1:
        raise ValueError("Need sources")
    cmd += "--sources "
    for source in sources:
        cmd += "'%s' " % source
    cmd += "--target '%s'" % target
    return cmd

def _run_command(cmd):
    try:
        raw_output = subprocess.check_output(cmd, shell=True, bufsize=4096,
                                             stderr=subprocess.STDOUT)
        exit_code = 0
    except subprocess.CalledProcessError as cpe:
        raw_output = cpe.output
        exit_code = cpe.returncode
    return exit_code, raw_output.decode("UTF-8")

def _relative_path(full, base):
    return str(PosixPath(full).absolute().relative_to(PosixPath(base)))


def prepare_data_move_operation(sources, target, mount_info):
    """ Create command to be executed for a data copy operation.
        Will resolve source/target endpoint, and determine what operation 
        (GET or PUT) is needed.
        Will authenticate with the stored credentials.
    """
    _target_mount = _resolve_mount_point(target, mount_info)
    _source_mount = _resolve_mount_point(sources[0], mount_info)
    if _target_mount!=None:
        _method = "PUT"
        _mount = _target_mount
        _sources = sources
        _target = _relative_path(target, _mount["mount_point"])
    else:
        _method = "GET"
        _mount = _source_mount
        _target = target
        _sources = [ _relative_path(full, _mount["mount_point"]) for full in sources ]

    _auth_url = _mount["endpoint"]
    _credentials = _mount["credentials"]
    _remote_dir = _mount["remote_directory"]
    _uftp = create_uftp_handler(_remote_dir, _auth_url, _credentials)
    _host, _port, _pwd = _uftp.authenticate()

    return _create_command(_host, int(_port), _pwd, _method, _sources, _target)

def mount(mount_directory, parameters):
    """
    Authenticates to UFTPD and mounts the requested directory
    """
    auth_url = parameters['endpoint']
    remote_directory = parameters['remote_directory']
    mount_point = parameters['mount_point']
    credentials = parameters.get('credentials', {})
    try:
        uftp = create_uftp_handler(remote_directory, auth_url, credentials)
        (host, port, pwd) = uftp.authenticate()
        uftp.open_uftp_session(host, port, pwd)
        # avoid launching FUSE driver after an error during connect
        st = uftp.stat(".")
        error_code, output = run_fusedriver(host, port, pwd, mount_point)
    except Exception as ex:
        uftp.close()
        error_code = 1
        output = repr(ex)
    return error_code, output


def unmount(parameters):
    """
    Unmounts the requested directory
    """
    try:
        cmd = "fusermount -u '%s'" % parameters['mount_point']
        error_code, output = _run_command(cmd)
    except Exception as ex:
        error_code = 1
        output = repr(ex)
    return error_code, output


if __name__ == "__main__":
    """ 
    Main function to run download or upload from UFTPD.
    Requires the one-time password from authentication, which must have been done previously
    """
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("address", help="UFTPD server's address (host:port)")
    parser.add_argument("-P", "--password",
        help="one-time password (if not given, it is expected in the environment UFTP_PASSWORD)",
    )
    parser.add_argument("-X", "--operation", required=True, help="what to do, GET or PUT")
    parser.add_argument(
        "-s",
        "--sources",
        nargs="+",
        help="list of source files")
    parser.add_argument(
        "-t",
        "--target",
        help="target file or directory")

    args = parser.parse_args()
    _host, _port = args.address.split(":")
    _pwd = args.password
    if _pwd is None:
        _pwd = getenv("UFTP_PASSWORD")
    if _pwd is None:
        raise TypeError("UFTP one-time password must be given via --P or as environment UFTP_PASSWORD")
    _operation = args.operation
    if "GET"!=_operation and "PUT"!=_operation:
        raise TypeError("Not understood: %s" % _operation)

    uftp_session = UFTP()
    uftp_session.open_uftp_session(_host, int(_port), _pwd)

    if "GET"==_operation:
        _target = Path(args.target)
        if len(args.sources)>1:
            if not _target.is_dir():
                raise TypeError("target must be a directory")
        for source in args.sources:
            _source = PurePosixPath(source)
            if _target.is_dir():
                _write_to = _target / _source.name
            else:
                _write_to = _source
            with _write_to.open("wb") as fp:
                uftp_session.ftp.retrbinary('RETR %s' % str(_source), fp.write)
    elif "PUT"==_operation:
        _target = PurePosixPath(args.target)
        is_dir = len(args.sources)>1 or args.target.endswith("/")
        for source in args.sources:
            _source = Path(source)
            if is_dir:
                _write_to = _target / _source.name
            else:
                _write_to = _target
            with _source.open("rb") as fp:
                uftp_session.ftp.storbinary('STOR %s' % str(_write_to), fp)



from os import waitpid, WNOHANG
import subprocess
import threading
import time

class TaskHolder():
    """
    Holds info running asynchronous tasks. Starts a thread that tracks
    child processes and cleans up after they have finished
    """
    def __init__(self):
        self.tasks = []
        self.lock = threading.Lock()
        self.cleanup_thread = threading.Thread(target=self.cleanup,
                              name="Cleanup",
                              daemon=True,
                              args=())
        self.cleanup_thread.start()
    
    def add(self, task):
        self.tasks.append(task)
        
    def cleanup(self):
        while True:
            for task in self.tasks:
                if task.status == "FINISHED":
                    continue
                try:
                    (_pid, _status) = waitpid(task.child.pid, WNOHANG)
                    if _pid!=0:
                        task.status = "FINISHED"
                except ChildProcessError as cpe:
                    print(cpe)
            time.sleep(5)


class Task():
    """
    Holds info about a single asynchronous task, e.g. a long-running download operation
    """
    
    def __init__(self, cmd):
        self.child = None
        self.cmd = cmd
        self.status = "OK"

    def json(self):
        return { "pid": self.child.pid, 
                 "command": self.cmd,
                 "status": self.status
                }

    def launch(self):
        """
        Runs command
        """
        self.child = subprocess.Popen(self.cmd, shell=True)
        
